
using System;

namespace MiniClousot {



public class Parser {
	public const int _EOF = 0;
	public const int _ident = 1;
	public const int _number = 2;
	public const int maxT = 27;

	const bool T = true;
	const bool x = false;
	const int minErrDist = 2;
	
	public Scanner scanner;
	public Errors  errors;

	public Token t;    // last recognized token
	public Token la;   // lookahead token
	int errDist = minErrDist;

public Statement program;

/*--------------------------------------------------------------------------*/


	public Parser(Scanner scanner) {
		this.scanner = scanner;
		errors = new Errors();
	}

	void SynErr (int n) {
		if (errDist >= minErrDist) errors.SynErr(la.line, la.col, n);
		errDist = 0;
	}

	public void SemErr (string msg) {
		if (errDist >= minErrDist) errors.SemErr(t.line, t.col, msg);
		errDist = 0;
	}
	
	void Get () {
		for (;;) {
			t = la;
			la = scanner.Scan();
			if (la.kind <= maxT) { ++errDist; break; }

			la = t;
		}
	}
	
	void Expect (int n) {
		if (la.kind==n) Get(); else { SynErr(n); }
	}
	
	bool StartOf (int s) {
		return set[s, la.kind];
	}
	
	void ExpectWeak (int n, int follow) {
		if (la.kind == n) Get();
		else {
			SynErr(n);
			while (!StartOf(follow)) Get();
		}
	}


	bool WeakSeparator(int n, int syFol, int repFol) {
		int kind = la.kind;
		if (kind == n) {Get(); return true;}
		else if (StartOf(repFol)) {return false;}
		else {
			SynErr(n);
			while (!(set[syFol, kind] || set[repFol, kind] || set[0, kind])) {
				Get();
				kind = la.kind;
			}
			return StartOf(syFol);
		}
	}

	
	void AddOp(out BinaryOp op) {
		op = 0; 
		op = BinaryOp.Add; 
		if (la.kind == 3) {
			Get();
		} else if (la.kind == 4) {
			Get();
			op = BinaryOp.Sub; 
		} else SynErr(28);
	}

	void Expr(out Expression res) {
		BinaryOp op; Expression exp1, exp2; 
		SimExpr(out exp1);
		res = exp1; 
		if (StartOf(1)) {
			RelOp(out op);
			SimExpr(out exp2);
			res = new BinaryExpression(op, exp1, exp2); 
		}
	}

	void SimExpr(out Expression res) {
		BinaryOp op; Expression exp; 
		Term(out exp);
		res = exp; 
		while (la.kind == 3 || la.kind == 4) {
			AddOp(out op);
			Term(out exp);
			res = new BinaryExpression(op, res, exp); 
		}
	}

	void RelOp(out BinaryOp op) {
		op = 0; 
		switch (la.kind) {
		case 9: {
			Get();
			op = BinaryOp.Equal; 
			break;
		}
		case 10: {
			Get();
			op = BinaryOp.LessEqual; 
			break;
		}
		case 11: {
			Get();
			op = BinaryOp.Less; 
			break;
		}
		case 12: {
			Get();
			op = BinaryOp.GreaterEqual; 
			break;
		}
		case 13: {
			Get();
			op = BinaryOp.Greater; 
			break;
		}
		case 14: {
			Get();
			op = BinaryOp.NotEqual; 
			break;
		}
		default: SynErr(29); break;
		}
	}

	void Factor(out Expression res) {
		string name; Expression exp; res = null; 
		if (la.kind == 1) {
			Ident(out name);
			res = new Var(name); 
		} else if (la.kind == 2) {
			Get();
			int n = Convert.ToInt32(t.val); res = new Const(n); 
		} else if (la.kind == 4) {
			Get();
			Factor(out exp);
			res = new UnaryExpression(UnaryOp.Minus, exp); 
		} else if (la.kind == 5) {
			Get();
			
		} else if (la.kind == 6) {
			Get();
			
		} else SynErr(30);
	}

	void Ident(out string name) {
		Expect(1);
		name = t.val; 
	}

	void MulOp(out BinaryOp op) {
		op = BinaryOp.Mul; 
		if (la.kind == 7) {
			Get();
		} else if (la.kind == 8) {
			Get();
			op = BinaryOp.Div; 
		} else SynErr(31);
	}

	void Term(out Expression res) {
		Expression exp1, exp2; BinaryOp op; 
		Factor(out exp1);
		res = exp1; 
		while (la.kind == 7 || la.kind == 8) {
			MulOp(out op);
			Factor(out exp1);
			res = new BinaryExpression(op, exp1, res); 
		}
	}

	void Stat(out Statement res) {
		Statement s1, s2; Expression cond, exp1, exp2; string name; Var var;  res  = null; 
		switch (la.kind) {
		case 1: {
			Ident(out name);
			var = new Var(name); 
			Expect(15);
			Expr(out exp2);
			Expect(16);
			res = new Assign(var, exp2); 
			break;
		}
		case 17: {
			Get();
			Expect(18);
			Expr(out cond);
			Expect(19);
			Stat(out s1);
			res = new If(cond, s1, new Nop()); 
			if (la.kind == 20) {
				Get();
				Stat(out s2);
				res = new If(cond, s1, s2); 
			}
			break;
		}
		case 21: {
			Get();
			Expect(18);
			Expr(out cond);
			Expect(19);
			Stat(out s1);
			res = new While(cond, s1); 
			break;
		}
		case 22: {
			Get();
			Expr(out cond);
			Expect(16);
			res = new Assert(cond); 
			break;
		}
		case 23: {
			Get();
			Expr(out cond);
			Expect(16);
			res = new Assume(cond); 
			break;
		}
		case 24: {
			Get();
			Stat(out s1);
			res = s1; 
			while (StartOf(2)) {
				Stat(out s1);
				res = new Seq(res, s1); 
			}
			Expect(25);
			break;
		}
		default: SynErr(32); break;
		}
	}

	void MiniClousot() {
		string name;  
		Expect(26);
		Ident(out name);
		Console.WriteLine("Parsing {0}...", name); 
		Stat(out program);
	}



	public void Parse() {
		la = new Token();
		la.val = "";		
		Get();
		MiniClousot();

    Expect(0);
	}
	
	static readonly bool[,] set = {
		{T,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x},
		{x,x,x,x, x,x,x,x, x,T,T,T, T,T,T,x, x,x,x,x, x,x,x,x, x,x,x,x, x},
		{x,T,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,T,x,x, x,T,T,T, T,x,x,x, x}

	};
} // end Parser


public class Errors {
	public int count = 0;                                    // number of errors detected
	public System.IO.TextWriter errorStream = Console.Out;   // error messages go to this stream
  public string errMsgFormat = "-- line {0} col {1}: {2}"; // 0=line, 1=column, 2=text
  
	public void SynErr (int line, int col, int n) {
		string s;
		switch (n) {
			case 0: s = "EOF expected"; break;
			case 1: s = "ident expected"; break;
			case 2: s = "number expected"; break;
			case 3: s = "\"+\" expected"; break;
			case 4: s = "\"-\" expected"; break;
			case 5: s = "\"true\" expected"; break;
			case 6: s = "\"false\" expected"; break;
			case 7: s = "\"*\" expected"; break;
			case 8: s = "\"/\" expected"; break;
			case 9: s = "\"==\" expected"; break;
			case 10: s = "\"<=\" expected"; break;
			case 11: s = "\"<\" expected"; break;
			case 12: s = "\">=\" expected"; break;
			case 13: s = "\">\" expected"; break;
			case 14: s = "\"!=\" expected"; break;
			case 15: s = "\"=\" expected"; break;
			case 16: s = "\";\" expected"; break;
			case 17: s = "\"if\" expected"; break;
			case 18: s = "\"(\" expected"; break;
			case 19: s = "\")\" expected"; break;
			case 20: s = "\"else\" expected"; break;
			case 21: s = "\"while\" expected"; break;
			case 22: s = "\"assert\" expected"; break;
			case 23: s = "\"assume\" expected"; break;
			case 24: s = "\"{\" expected"; break;
			case 25: s = "\"}\" expected"; break;
			case 26: s = "\"program\" expected"; break;
			case 27: s = "??? expected"; break;
			case 28: s = "invalid AddOp"; break;
			case 29: s = "invalid RelOp"; break;
			case 30: s = "invalid Factor"; break;
			case 31: s = "invalid MulOp"; break;
			case 32: s = "invalid Stat"; break;

			default: s = "error " + n; break;
		}
		errorStream.WriteLine(errMsgFormat, line, col, s);
		count++;
	}

	public void SemErr (int line, int col, string s) {
		errorStream.WriteLine(errMsgFormat, line, col, s);
		count++;
	}
	
	public void SemErr (string s) {
		errorStream.WriteLine(s);
		count++;
	}
	
	public void Warning (int line, int col, string s) {
		errorStream.WriteLine(errMsgFormat, line, col, s);
	}
	
	public void Warning(string s) {
		errorStream.WriteLine(s);
	}
} // Errors


public class FatalError: Exception {
	public FatalError(string m): base(m) {}
}

}